# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from enum import Enum
from msrest.exceptions import ValidationError
from knack.log import get_logger
from knack.util import CLIError
from azure.cli.core.commands import LongRunningOperation
from azure.cli.core.commands.client_factory import get_subscription_id
from ._client_factory import cf_acr_tokens, cf_acr_scope_maps
from ._utils import (
    build_token_id,
    create_default_scope_map,
    get_registry_by_name,
    get_scope_map_from_id,
    get_token_from_id,
    parse_scope_map_actions,
    user_confirmation,
    validate_managed_registry
)


class ConnectedRegistryModes(Enum):
    MIRROR = 'Mirror'
    REGISTRY = 'Registry'


class ConnectedRegistryActivationStatus(Enum):
    ACTIVE = 'Active'
    INACTIVE = 'Inactive'


DEFAULT_GATEWAY_SCOPE = ['config/read', 'config/write', 'message/read', 'message/write']
REPO_SCOPES_BY_MODE = {
    ConnectedRegistryModes.MIRROR.value: ['content/read', 'metadata/read'],
    ConnectedRegistryModes.REGISTRY.value: ['content/read', 'content/write', 'content/delete',
                                            'metadata/read', 'metadata/write']
}
SYNC_SCOPE_MAP_NAME = "{}-sync-scope-map"
SYNC_TOKEN_NAME = "{}-sync-token"
REPOSITORY = "repositories/"
GATEWAY = "gateway/"

logger = get_logger(__name__)


def acr_connected_registry_create(cmd,  # pylint: disable=too-many-locals, too-many-statements
                                  client,
                                  registry_name,
                                  connected_registry_name,
                                  repositories=None,
                                  sync_token_name=None,
                                  client_token_list=None,
                                  resource_group_name=None,
                                  mode=None,
                                  parent_name=None,
                                  sync_schedule=None,
                                  sync_message_ttl=None,
                                  sync_window=None,
                                  log_level=None,
                                  sync_audit_logs_enabled=False):

    if bool(sync_token_name) == bool(repositories):
        raise CLIError("usage error: you need to provide either --sync-token-name or --repository, but not both.")
    registry, resource_group_name = get_registry_by_name(cmd.cli_ctx, registry_name, resource_group_name)
    subscription_id = get_subscription_id(cmd.cli_ctx)

    if not registry.data_endpoint_enabled:
        raise CLIError("Can't create the connected registry '{}' ".format(connected_registry_name) +
                       "because the cloud registry '{}' data endpoint is disabled. ".format(registry_name) +
                       "Enabling the data endpoint might affect your firewall rules.\nTo enable data endpoint run:" +
                       "\n\taz acr update -n {} --data-endpoint-enabled true".format(registry_name))

    ErrorResponseException = cmd.get_models('ErrorResponseException')
    parent = None
    mode = mode.capitalize()
    if parent_name:
        try:
            parent = acr_connected_registry_show(cmd, client, parent_name, registry_name, resource_group_name)
            connected_registry_list = list(client.list(resource_group_name, registry_name))
            family_tree, _ = _get_family_tree(connected_registry_list, None)
        except ErrorResponseException as ex:
            if ex.response.status_code == 404:
                raise CLIError("The parent connected registry '{}' could not be found.".format(parent_name))
            raise CLIError(ex)
        if parent.mode != ConnectedRegistryModes.REGISTRY.value and parent.mode != mode:
            raise CLIError("Can't create the registry '{}' with mode '{}' ".format(connected_registry_name, mode) +
                           "when the connected registry parent '{}' mode is '{}'. ".format(parent_name, parent.mode) +
                           "For more information on connected registries " +
                           "please visit https://aka.ms/acr/connected-registry.")
        _update_ancestor_permissions(cmd, family_tree, resource_group_name, registry_name, parent.id,
                                     connected_registry_name, repositories, mode, False)

    if sync_token_name:
        sync_token_id = build_token_id(subscription_id, resource_group_name, registry_name, sync_token_name)
    else:
        sync_token_id = _create_sync_token(cmd, resource_group_name, registry_name,
                                           connected_registry_name, repositories, mode)

    if client_token_list is not None:
        for i, client_token_name in enumerate(client_token_list):
            client_token_list[i] = build_token_id(
                subscription_id, resource_group_name, registry_name, client_token_name)

    ConnectedRegistry, LoggingProperties, SyncProperties, ParentProperties = cmd.get_models(
        'ConnectedRegistry', 'LoggingProperties', 'SyncProperties', 'ParentProperties')
    connected_registry_create_parameters = ConnectedRegistry(
        provisioning_state=None,
        mode=mode,
        parent=ParentProperties(
            id=parent.id if parent else None,
            sync_properties=SyncProperties(
                token_id=sync_token_id,
                schedule=sync_schedule,
                message_ttl=sync_message_ttl,
                sync_window=sync_window
            )
        ),
        client_token_ids=client_token_list,
        logging=LoggingProperties(
            log_level=log_level,
            audit_log_status='Enabled' if sync_audit_logs_enabled else 'Disabled'
        )
    )

    try:
        return client.create(subscription_id=subscription_id,
                             resource_group_name=resource_group_name,
                             registry_name=registry_name,
                             connected_registry_name=connected_registry_name,
                             connected_registry_create_parameters=connected_registry_create_parameters)
    except ValidationError as e:
        raise CLIError(e)


def acr_connected_registry_update(cmd,  # pylint: disable=too-many-locals, too-many-statements
                                  client,
                                  registry_name,
                                  connected_registry_name,
                                  add_client_token_list=None,
                                  remove_client_token_list=None,
                                  resource_group_name=None,
                                  sync_schedule=None,
                                  sync_window=None,
                                  log_level=None,
                                  sync_message_ttl=None,
                                  sync_audit_logs_enabled=None):
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)
    subscription_id = get_subscription_id(cmd.cli_ctx)
    current_connected_registry = acr_connected_registry_show(
        cmd, client, connected_registry_name, registry_name, resource_group_name)

    # Add or remove from the current client token id list
    if add_client_token_list is not None:
        for i, client_token_name in enumerate(add_client_token_list):
            add_client_token_list[i] = build_token_id(
                subscription_id, resource_group_name, registry_name, client_token_name)
        add_client_token_set = set(add_client_token_list)
    else:
        add_client_token_set = set()
    if remove_client_token_list is not None:
        for i, client_token_name in enumerate(remove_client_token_list):
            remove_client_token_list[i] = build_token_id(
                subscription_id, resource_group_name, registry_name, client_token_name)
        remove_client_token_set = set(remove_client_token_list)
    else:
        remove_client_token_set = set()

    duplicate_client_token = set.intersection(add_client_token_set, remove_client_token_set)
    if duplicate_client_token:
        errors = sorted(map(lambda action: action[action.rfind('/') + 1:], duplicate_client_token))
        raise CLIError(
            'Update ambiguity. Duplicate client token ids were provided with ' +
            '--add-client-tokens and --remove-client-tokens arguments.\n{}'.format(errors))

    current_client_token_set = set(current_connected_registry.client_token_ids) \
        if current_connected_registry.client_token_ids else set()
    client_token_set = current_client_token_set.union(add_client_token_set).difference(remove_client_token_set)

    client_token_list = list(client_token_set) if client_token_set != current_client_token_set else None

    ConnectedRegistryUpdateParameters, SyncUpdateProperties, LoggingProperties = cmd.get_models(
        'ConnectedRegistryUpdateParameters', 'SyncUpdateProperties', 'LoggingProperties')
    connected_registry_update_parameters = ConnectedRegistryUpdateParameters(
        sync_properties=SyncUpdateProperties(
            token_id=current_connected_registry.parent.sync_properties.token_id,
            schedule=sync_schedule,
            message_ttl=sync_message_ttl,
            sync_window=sync_window
        ),
        logging=LoggingProperties(
            log_level=log_level,
            audit_log_status=sync_audit_logs_enabled
        ),
        client_token_ids=client_token_list
    )

    try:
        return client.update(resource_group_name=resource_group_name,
                             registry_name=registry_name,
                             connected_registry_name=connected_registry_name,
                             connected_registry_update_parameters=connected_registry_update_parameters)
    except ValidationError as e:
        raise CLIError(e)


def acr_connected_registry_delete(cmd,
                                  client,
                                  connected_registry_name,
                                  registry_name,
                                  cleanup=False,
                                  yes=False,
                                  resource_group_name=None):
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)
    user_confirmation("Are you sure you want to delete the connected registry '{}' in '{}'?".format(
        connected_registry_name, registry_name), yes)
    try:
        connected_registry = acr_connected_registry_show(
            cmd, client, connected_registry_name, registry_name, resource_group_name)
        result = client.delete(resource_group_name, registry_name, connected_registry_name)
        sync_token = get_token_from_id(cmd, connected_registry.parent.sync_properties.token_id)
        sync_token_name = sync_token.name
        sync_scope_map_name = sync_token.scope_map_id.split('/scopeMaps/')[1]
        if cleanup:
            from .token import acr_token_delete
            from .scope_map import acr_scope_map_delete
            token_client = cf_acr_tokens(cmd.cli_ctx)
            scope_map_client = cf_acr_scope_maps(cmd.cli_ctx)

            # Delete target sync scope map and token.
            acr_token_delete(cmd, token_client, registry_name, sync_token_name, yes, resource_group_name)
            acr_scope_map_delete(cmd, scope_map_client, registry_name, sync_scope_map_name, yes, resource_group_name)
            # Cleanup gateway permissions from ancestors
            connected_registry_list = list(client.list(resource_group_name, registry_name))
            family_tree, _ = _get_family_tree(connected_registry_list, None)
            _update_ancestor_permissions(cmd, family_tree, resource_group_name, registry_name,
                                         connected_registry.parent.id, connected_registry_name, remove_access=True)
        else:
            msg = "Connected registry successfully deleted. Please cleanup your sync tokens and scope maps. " + \
                "Run the following commands for cleanup: \n\t" + \
                "az acr token delete -n {} -r {} --yes\n\t".format(sync_token_name, registry_name) + \
                "az acr scope-map delete -n {} -r {} --yes\n".format(sync_scope_map_name, registry_name) + \
                "Run the following command on all ascendency to remove the deleted registry gateway access: \n\t" + \
                "az acr scope-map update -n <scope-map-name> -r {} --remove-gateway {} --yes".format(
                    registry_name, " ".join([connected_registry_name] + DEFAULT_GATEWAY_SCOPE))
            logger.warning(msg)
        return result
    except ValidationError as e:
        raise CLIError(e)


def acr_connected_registry_deactivate(cmd,
                                      client,
                                      connected_registry_name,
                                      registry_name,
                                      yes=False,
                                      resource_group_name=None):
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)
    subscription_id = get_subscription_id(cmd.cli_ctx)

    user_confirmation("Are you sure you want to deactivate the connected registry '{}' in '{}'?".format(
        connected_registry_name, registry_name), yes)
    return client.deactivate(subscription_id=subscription_id,
                             resource_group_name=resource_group_name,
                             registry_name=registry_name,
                             connected_registry_name=connected_registry_name)


def acr_connected_registry_list(cmd,
                                client,
                                registry_name,
                                parent_name=None,
                                no_children=False,
                                resource_group_name=None):
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)
    connected_registry_list = list(client.list(resource_group_name, registry_name))
    result = []
    if no_children:
        if parent_name:
            result = [registry for registry in connected_registry_list
                      if registry.parent.id is not None and registry.parent.id.endswith(parent_name)]
        else:
            result = [registry for registry in connected_registry_list if not registry.parent.id]
    elif parent_name:
        family_tree, parent = _get_family_tree(connected_registry_list, parent_name)
        if parent is None:
            raise CLIError("Parent connected registry '{}' doesn't exist.".format(parent_name))
        result = _get_descendants(family_tree, parent.id)
    else:
        result = connected_registry_list
    return result


def acr_connected_registry_show(cmd,
                                client,
                                connected_registry_name,
                                registry_name,
                                resource_group_name=None):
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)
    return client.get(resource_group_name, registry_name, connected_registry_name)


def acr_connected_registry_list_client_tokens(cmd,
                                              client,
                                              connected_registry_name,
                                              registry_name,
                                              resource_group_name=None):
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)
    current_connected_registry = acr_connected_registry_show(
        cmd, client, connected_registry_name, registry_name, resource_group_name)

    result = []
    if current_connected_registry.client_token_ids is None:
        return result

    for token_id in current_connected_registry.client_token_ids:
        token = get_token_from_id(cmd, token_id)
        result.append(token)
    return result


def _create_sync_token(cmd,
                       resource_group_name,
                       registry_name,
                       connected_registry_name,
                       repositories,
                       mode):
    token_client = cf_acr_tokens(cmd.cli_ctx)

    mode = mode.capitalize()
    if not any(option for option in ConnectedRegistryModes if option.value == mode):
        raise CLIError("usage error: --mode supports only 'registry' and 'mirror' values.")
    repository_actions_list = [[repo] + REPO_SCOPES_BY_MODE[mode] for repo in repositories]
    gateway_actions_list = [[connected_registry_name.lower()] + DEFAULT_GATEWAY_SCOPE]
    try:
        message = "Created by connected registry sync token: {}"
        sync_scope_map_name = SYNC_SCOPE_MAP_NAME.format(connected_registry_name)
        logger.warning("If sync scope map '%s' already exists, its actions will be overwritten", sync_scope_map_name)
        sync_scope_map = create_default_scope_map(cmd, resource_group_name, registry_name, sync_scope_map_name,
                                                  repository_actions_list, gateway_actions_list,
                                                  scope_map_description=message.format(connected_registry_name),
                                                  force=True)

        sync_token_name = SYNC_TOKEN_NAME.format(connected_registry_name)
        logger.warning("If sync token '%s' already exists, it properties will be overwritten", sync_token_name)
        Token = cmd.get_models('Token')
        poller = token_client.create(
            resource_group_name,
            registry_name,
            sync_token_name,
            Token(
                scope_map_id=sync_scope_map.id,
                status="enabled"
            )
        )

        token = LongRunningOperation(cmd.cli_ctx)(poller)
        return token.id
    except ValidationError as e:
        raise CLIError(e)


def _get_family_tree(connected_registry_list, target_connected_registry_name):
    family_tree = {}
    targetConnectedRegistry = None
    # Populate the dictionary
    for ConnectedRegistry in connected_registry_list:
        family_tree[ConnectedRegistry.id] = {
            "connectedRegistry": ConnectedRegistry,
            "children": []
        }
        if ConnectedRegistry.name == target_connected_registry_name:
            targetConnectedRegistry = ConnectedRegistry

    # Populate Children dependencies
    for ConnectedRegistry in connected_registry_list:
        parent_id = ConnectedRegistry.parent.id
        if parent_id and not parent_id.isspace():
            family_tree[parent_id]["children"].append(ConnectedRegistry.id)
    return family_tree, targetConnectedRegistry


def _get_descendants(family_tree, parent_id):
    children = family_tree[parent_id]['children']
    result = []
    for child_id in children:
        result = [family_tree[child_id]["connectedRegistry"]]
        descendants = _get_descendants(family_tree, child_id)
        if descendants:
            result.extend(descendants)
    return result


# region connected-registry install subgroup
def acr_connected_registry_install_info(cmd,
                                        client,
                                        connected_registry_name,
                                        registry_name,
                                        resource_group_name=None):
    return _get_install_info(cmd, client, connected_registry_name, registry_name, False, resource_group_name)


def acr_connected_registry_install_renew_credentials(cmd,
                                                     client,
                                                     connected_registry_name,
                                                     registry_name,
                                                     resource_group_name=None):
    return _get_install_info(cmd, client, connected_registry_name, registry_name, True, resource_group_name)


def _get_install_info(cmd,
                      client,
                      connected_registry_name,
                      registry_name,
                      regenerate_credentials,
                      resource_group_name=None):
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)
    connected_registry = acr_connected_registry_show(
        cmd, client, connected_registry_name, registry_name, resource_group_name)
    parent_gateway_endpoint = connected_registry.parent.sync_properties.gateway_endpoint
    if parent_gateway_endpoint is None or parent_gateway_endpoint == '':
        parent_gateway_endpoint = "<parent gateway endpoint>"
    parent_id = connected_registry.parent.id
    # if parent_id is not none, parent is a connected registry
    if parent_id:
        parent_endpoint_protocol = "<http or https>"
    # if parent_id is none, parent is a cloud registry
    else:
        parent_endpoint_protocol = "https"
    sync_token_name = connected_registry.parent.sync_properties.token_id.split('/tokens/')[1]

    connected_registry_login_server = "<Optional: connected registry login server. " + \
        "More info at https://aka.ms/acr/connected-registry>"

    if regenerate_credentials:
        from ._client_factory import cf_acr_token_credentials
        from .token import acr_token_credential_generate
        cred_client = cf_acr_token_credentials(cmd.cli_ctx)
        poller = acr_token_credential_generate(
            cmd, cred_client, registry_name, sync_token_name,
            password1=True, password2=False, resource_group_name=resource_group_name)
        credentials = LongRunningOperation(cmd.cli_ctx)(poller)
        sync_username = credentials.username
        sync_password = credentials.passwords[0].value
        logger.warning('Please store your generated credentials safely.')
    else:
        sync_username = sync_token_name
        sync_password = "<sync token password>"

    connection_string = "ConnectedRegistryName=%s;" % connected_registry_name + \
        "SyncTokenName=%s;SyncTokenPassword=%s;" % (sync_username, sync_password) + \
        "ParentGatewayEndpoint=%s;ParentEndpointProtocol=%s" % (parent_gateway_endpoint, parent_endpoint_protocol)
    return {
        "ACR_REGISTRY_CONNECTION_STRING": connection_string,
        "ACR_REGISTRY_LOGIN_SERVER": connected_registry_login_server
    }
# endregion


def _update_ancestor_permissions(cmd,
                                 family_tree,
                                 resource_group_name,
                                 registry_name,
                                 parent_id,
                                 gateway,
                                 repositories=None,
                                 mode=None,
                                 remove_access=False):
    gateway_actions_list = [[gateway.lower()] + DEFAULT_GATEWAY_SCOPE]
    if repositories is not None:
        repository_actions_list = [[repo] + REPO_SCOPES_BY_MODE[mode] for repo in repositories]
        repo_msg = ", ".join(repositories)
        repo_msg = " and repo(s) '{}' {} permissions".format(repo_msg, mode)
    if remove_access:
        action_txt = "Removing"
        add_actions_set = set()
        remove_actions_set = set(parse_scope_map_actions(gateway_actions_list=gateway_actions_list))
    else:
        action_txt = "Adding"
        add_actions_set = set(parse_scope_map_actions(repository_actions_list, gateway_actions_list))
        remove_actions_set = set()

    while parent_id and not parent_id.isspace():
        ancestor = family_tree[parent_id]["connectedRegistry"]
        msg = "{} '{}' gateway permissions{} to connected registry '{}' sync scope map.".format(
            action_txt, gateway, repo_msg, ancestor.name)
        _update_repo_permissions(cmd, resource_group_name, registry_name,
                                 ancestor, add_actions_set, remove_actions_set, msg=msg)
        parent_id = ancestor.parent.id


# region connected-registry repo update
def _update_repo_permissions(cmd,
                             resource_group_name,
                             registry_name,
                             connected_registry,
                             add_actions_set,
                             remove_actions_set,
                             msg=None,
                             description=None):
    scope_map_client = cf_acr_scope_maps(cmd.cli_ctx)
    sync_token = get_token_from_id(cmd, connected_registry.parent.sync_properties.token_id)
    sync_scope_map = get_scope_map_from_id(cmd, sync_token.scope_map_id)
    sync_scope_map_name = sync_scope_map.name
    current_actions_set = set(sync_scope_map.actions)
    final_actions_set = current_actions_set.union(add_actions_set).difference(remove_actions_set)
    if final_actions_set == current_actions_set:
        return None
    current_actions = list(final_actions_set)
    logger.warning(msg)
    return scope_map_client.update(
        resource_group_name,
        registry_name,
        sync_scope_map_name,
        description,
        current_actions
    )


def _get_scope_map_actions_set(repos, actions):
    for i, repo_name in enumerate(repos):
        repos[i] = [repo_name] + actions
    return set(parse_scope_map_actions(repos))


def acr_connected_registry_repo(cmd,
                                client,
                                connected_registry_name,
                                registry_name,
                                add_repos=None,
                                remove_repos=None,
                                resource_group_name=None):
    if not (add_repos or remove_repos):
        raise CLIError('No repository permissions to update.')
    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name)

    add_repos_set = set(add_repos) if add_repos is not None else set()
    remove_repos_set = set(remove_repos) if remove_repos is not None else set()
    duplicate_repos = set.intersection(add_repos_set, remove_repos_set)
    if duplicate_repos:
        errors = sorted(map(lambda action: action[action.rfind('/') + 1:], duplicate_repos))
        raise CLIError(
            'Update ambiguity. Duplicate repository names were provided with ' +
            '--add and --remove arguments.\n{}'.format(errors))

    connected_registry_list = list(client.list(resource_group_name, registry_name))
    family_tree, target_connected_registry = _get_family_tree(connected_registry_list, connected_registry_name)
    if target_connected_registry is None:
        raise CLIError("Connected registry '{}' doesn't exist.".format(connected_registry_name))

    # remove repo permissions from connected registry descendants.
    remove_actions = REPO_SCOPES_BY_MODE[ConnectedRegistryModes.REGISTRY.value]
    if remove_repos is not None:
        remove_repos_txt = ", ".join(remove_repos)
        remove_repos_set = _get_scope_map_actions_set(remove_repos, remove_actions)
        descendants = _get_descendants(family_tree, target_connected_registry.id)
        for connected_registry in descendants:
            msg = "Removing '{}' permissions from {}".format(remove_repos_txt, connected_registry.name)
            _update_repo_permissions(cmd, resource_group_name, registry_name,
                                     connected_registry, set(), remove_repos_set, msg=msg)
    else:
        remove_repos_set = set()

    # add repo permissions to ancestors.
    add_actions = REPO_SCOPES_BY_MODE[target_connected_registry.mode]
    if add_repos is not None:
        add_repos_txt = ", ".join(add_repos)
        add_repos_set = _get_scope_map_actions_set(add_repos, add_actions)
        parent_id = target_connected_registry.parent.id
        while parent_id and not parent_id.isspace():
            connected_registry = family_tree[parent_id]["connectedRegistry"]
            msg = "Adding '{}' permissions to {}".format(add_repos_txt, connected_registry.name)
            _update_repo_permissions(cmd, resource_group_name, registry_name,
                                     connected_registry, add_repos_set, set(), msg=msg)
            parent_id = connected_registry.parent.id
    else:
        add_repos_set = set()

    # update target connected registry repo permissions.
    if add_repos and remove_repos:
        msg = "Adding '{}' and removing '{}' permissions in {}".format(
            add_repos_txt, remove_repos_txt, target_connected_registry.name)
    elif add_repos:
        msg = "Adding '{}' permissions to {}".format(add_repos_txt, target_connected_registry.name)
    else:
        msg = "Removing '{}' permissions from {}".format(remove_repos_txt, target_connected_registry.name)
    _update_repo_permissions(cmd, resource_group_name, registry_name,
                             target_connected_registry, add_repos_set, remove_repos_set, msg=msg)
# endregion



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands import CliCommandType
from azure.cli.command_modules.backup._client_factory import vaults_cf, backup_protection_containers_cf, \
    protection_policies_cf, backup_policies_cf, protected_items_cf, backups_cf, backup_jobs_cf, \
    job_details_cf, job_cancellations_cf, recovery_points_cf, restores_cf, backup_storage_configs_cf, \
    item_level_recovery_connections_cf, backup_protected_items_cf, backup_protectable_items_cf, \
    protection_containers_cf, protection_intent_cf  # pylint: disable=unused-variable
from azure.cli.command_modules.backup._format import (
    transform_container_list, transform_policy_list, transform_item_list, transform_job_list,
    transform_recovery_point_list, transform_container, transform_item, transform_protectable_item_list, transform_job,
    transform_log_chain_list)


# pylint: disable=line-too-long
# pylint: disable=too-many-statements
def load_command_table(self, _):

    backup_custom = CliCommandType(operations_tmpl='azure.cli.command_modules.backup.custom#{}')

    backup_custom_base = CliCommandType(operations_tmpl='azure.cli.command_modules.backup.custom_base#{}')

    backup_vaults_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.recoveryservices.operations#VaultsOperations.{}',
        client_factory=vaults_cf)

    with self.command_group('backup vault', backup_vaults_sdk, client_factory=vaults_cf) as g:
        g.custom_command('create', 'create_vault')
        g.show_command('show', 'get')
        g.custom_command('list', 'list_vaults')
        g.custom_command('backup-properties show', 'get_backup_properties', client_factory=backup_storage_configs_cf)
        g.custom_command('backup-properties set', 'set_backup_properties', client_factory=backup_storage_configs_cf)
        g.custom_command('delete', 'delete_vault', confirmation=True)

    with self.command_group('backup container', backup_custom_base, client_factory=protection_containers_cf) as g:
        g.show_command('show', 'show_container', client_factory=backup_protection_containers_cf, table_transformer=transform_container)
        g.command('list', 'list_containers', table_transformer=transform_container_list, client_factory=backup_protection_containers_cf)

    with self.command_group('backup container', custom_command_type=backup_custom_base, client_factory=protection_containers_cf) as g:
        g.custom_command('unregister', 'unregister_container', confirmation=True)
        g.custom_command('re-register', 're_register_wl_container', confirmation=True)
        g.custom_command('register', 'register_wl_container')

    with self.command_group('backup policy', backup_custom_base, client_factory=protection_policies_cf) as g:
        g.command('get-default-for-vm', 'get_default_policy_for_vm')
        g.show_command('show', 'show_policy')
        g.command('list', 'list_policies', client_factory=backup_policies_cf, table_transformer=transform_policy_list)
        g.command('list-associated-items', 'list_associated_items_for_policy', client_factory=backup_protected_items_cf, table_transformer=transform_item_list)

    with self.command_group('backup policy', custom_command_type=backup_custom_base, client_factory=protection_policies_cf) as g:
        g.custom_command('set', 'set_policy')
        g.custom_command('delete', 'delete_policy')
        g.custom_command('create', 'create_policy')

    with self.command_group('backup protection', backup_custom_base, client_factory=protected_items_cf) as g:
        g.command('check-vm', 'check_protection_enabled_for_vm')
        g.command('enable-for-vm', 'enable_protection_for_vm')
        g.command('update-for-vm', 'update_protection_for_vm')

    with self.command_group('backup protection', custom_command_type=backup_custom_base, client_factory=protected_items_cf) as g:
        g.custom_command('backup-now', 'backup_now', client_factory=backups_cf, table_transformer=transform_job)
        g.custom_command('disable', 'disable_protection', confirmation=True)
        g.custom_command('enable-for-azurefileshare', 'enable_for_azurefileshare')
        g.custom_command('enable-for-azurewl', 'enable_protection_for_azure_wl')
        g.custom_command('auto-enable-for-azurewl', 'auto_enable_for_azure_wl', client_factory=protection_intent_cf)
        g.custom_command('auto-disable-for-azurewl', 'disable_auto_for_azure_wl', client_factory=protection_intent_cf)
        g.custom_command('resume', 'resume_protection')
        g.custom_command('undelete', 'undelete_protection')

    with self.command_group('backup item', backup_custom_base, client_factory=protected_items_cf) as g:
        g.show_command('show', 'show_item', client_factory=backup_protected_items_cf, table_transformer=transform_item)
        g.command('list', 'list_items', table_transformer=transform_item_list, client_factory=backup_protected_items_cf)
        g.command('set-policy', 'update_policy_for_item', table_transformer=transform_job)

    with self.command_group('backup protectable-item', backup_custom_base, client_factory=backup_protectable_items_cf) as g:
        g.show_command('show', 'show_protectable_item')
        g.command('list', 'list_protectable_items', table_transformer=transform_protectable_item_list)
        g.command('initialize', 'initialize_protectable_items', client_factory=protection_containers_cf)

    with self.command_group('backup job', backup_custom, client_factory=job_details_cf) as g:
        g.command('list', 'list_jobs', client_factory=backup_jobs_cf, table_transformer=transform_job_list)
        g.show_command('show', 'show_job')
        g.command('stop', 'stop_job', client_factory=job_cancellations_cf)
        g.command('wait', 'wait_for_job')

    with self.command_group('backup recoverypoint', backup_custom_base, client_factory=recovery_points_cf) as g:
        g.show_command('show', 'show_recovery_point')
        g.command('list', 'list_recovery_points', table_transformer=transform_recovery_point_list)
        g.show_command('show-log-chain', 'list_recovery_points', table_transformer=transform_log_chain_list)

    with self.command_group('backup restore', backup_custom_base, client_factory=restores_cf) as g:
        g.command('restore-disks', 'restore_disks')

    with self.command_group('backup restore', custom_command_type=backup_custom_base, client_factory=restores_cf) as g:
        g.custom_command('restore-azurefileshare', 'restore_azurefileshare')
        g.custom_command('restore-azurefiles', 'restore_azurefiles')
        g.custom_command('restore-azurewl', 'restore_azure_wl', table_transformer=transform_job)

    with self.command_group('backup restore files', backup_custom, client_factory=item_level_recovery_connections_cf) as g:
        g.command('mount-rp', 'restore_files_mount_rp')
        g.command('unmount-rp', 'restore_files_unmount_rp')

    with self.command_group('backup recoveryconfig', backup_custom_base, client_factory=recovery_points_cf) as g:
        g.show_command('show', 'show_recovery_config')

    with self.command_group('backup', is_preview=True):
        pass



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.util import CLIError

from ._utils import get_registry_by_name


def acr_credential_show(cmd, client, registry_name, resource_group_name=None):
    registry, resource_group_name = get_registry_by_name(cmd.cli_ctx, registry_name, resource_group_name)

    if registry.admin_user_enabled:  # pylint: disable=no-member
        return client.list_credentials(resource_group_name, registry_name)

    raise admin_not_enabled_error(registry_name)


def acr_credential_renew(cmd, client, registry_name, password_name, resource_group_name=None):
    registry, resource_group_name = get_registry_by_name(cmd.cli_ctx, registry_name, resource_group_name)

    if registry.admin_user_enabled:  # pylint: disable=no-member
        return client.regenerate_credential(
            resource_group_name, registry_name, password_name)

    raise admin_not_enabled_error(registry_name)


def admin_not_enabled_error(registry_name):
    return CLIError("Run 'az acr update -n {} --admin-enabled true' to enable admin first.".format(registry_name))



# coding=utf-8
# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.help_files import helps  # pylint: disable=unused-import
# pylint: disable=line-too-long, too-many-lines

helps['policy event'] = """
type: group
short-summary: Manage policy events.
"""

helps['policy event list'] = """
type: command
short-summary: List policy events.
examples:
  - name: Get policy events at current subscription scope created in the last day.
    text: >
        az policy event list
  - name: Get policy events at management group scope.
    text: >
        az policy event list -m "myMg"
  - name: Get policy events at resource group scope in current subscription.
    text: >
        az policy event list -g "myRg"
  - name: Get policy events for a resource using resource ID.
    text: >
        az policy event list --resource "/subscriptions/fff10b27-fff3-fff5-fff8-fffbe01e86a5/resourceGroups/myResourceGroup /providers/Microsoft.EventHub/namespaces/myns1/eventhubs/eh1/consumergroups/cg1"
  - name: Get policy events for a resource using resource name.
    text: >
        az policy event list --resource "myKeyVault" --namespace "Microsoft.KeyVault" --resource-type "vaults" -g "myresourcegroup"
  - name: Get policy events for a nested resource using resource name.
    text: >
        az policy event list --resource "myRule1" --namespace "Microsoft.Network" --resource-type "securityRules" --parent "networkSecurityGroups/mysecuritygroup1" -g "myresourcegroup"
  - name: Get policy events for a policy set definition in current subscription.
    text: >
        az policy event list -s "fff58873-fff8-fff5-fffc-fffbe7c9d697"
  - name: Get policy events for a policy definition in current subscription.
    text: >
        az policy event list -d "fff69973-fff8-fff5-fffc-fffbe7c9d698"
  - name: Get policy events for a policy assignment in current subscription.
    text: >
        az policy event list -a "ddd8ef92e3714a5ea3d208c1"
  - name: Get policy events for a policy assignment in the specified resource group in current subscription.
    text: >
        az policy event list -g "myRg" -a "ddd8ef92e3714a5ea3d208c1"
  - name: Get top 5 policy events in current subscription, selecting a subset of properties and customizing ordering.
    text: >
        az policy event list --top 5 --order-by "timestamp desc, policyAssignmentName asc" --select "timestamp, resourceId, policyAssignmentId, policySetDefinitionId, policyDefinitionId"
  - name: Get policy events in current subscription during a custom time interval.
    text: >
        az policy event list --from "2018-03-08T00:00:00Z" --to "2018-03-15T00:00:00Z"
  - name: Get policy events in current subscription filtering results based on some property values.
    text: >
        az policy event list --filter "(policyDefinitionAction eq 'deny' or policyDefinitionAction eq 'audit') and resourceLocation ne 'eastus'"
  - name: Get number of policy events in current subscription.
    text: >
        az policy event list --apply "aggregate($count as numberOfRecords)"
  - name: Get policy events in current subscription aggregating results based on some properties.
    text: >
        az policy event list --apply "groupby((policyAssignmentId, policyDefinitionId, policyDefinitionAction, resourceId), aggregate($count as numEvents))"
  - name: Get policy events in current subscription grouping results based on some properties.
    text: >
        az policy event list --apply "groupby((policyAssignmentName, resourceId))"
  - name: Get policy events in current subscription aggregating results based on some properties specifying multiple groupings.
    text: >
        az policy event list --apply "groupby((policyAssignmentId, policyDefinitionId, resourceId))/groupby((policyAssignmentId, policyDefinitionId), aggregate($count as numResourcesWithEvents))"
"""

helps['policy remediation'] = """
type: group
short-summary: Manage resource policy remediations.
"""

helps['policy remediation cancel'] = """
type: command
short-summary: Cancel a resource policy remediation.
"""

helps['policy remediation create'] = """
type: command
short-summary: Create a resource policy remediation.
examples:
  - name: Create a remediation at resource group scope for a policy assignment
    text: >
        az policy remediation create -g myRg -n myRemediation --policy-assignment eeb18edc813c42d0ad5a9eab
  - name: Create a remediation at resource group scope for a policy assignment using the policy assignment resource ID
    text: >
        az policy remediation create -g myRg -n myRemediation --policy-assignment "/subscriptions/fff10b27-fff3-fff5-fff8-fffbe01e86a5/providers/Microsoft.Authorization/policyAssignments/myPa"
  - name: Create a remediation at subscription scope for a policy set assignment
    text: >
        az policy remediation create -n myRemediation --policy-assignment eeb18edc813c42d0ad5a9eab --definition-reference-id auditVMPolicyReference
  - name: Create a remediation at management group scope for specific resource locations
    text: >
        az policy remediation create -m myMg -n myRemediation --policy-assignment eeb18edc813c42d0ad5a9eab --location-filters eastus westeurope
  - name: Create a remediation for a specific resource using the resource ID
    text: >
        az policy remediation create --resource "/subscriptions/fff10b27-fff3-fff5-fff8-fffbe01e86a5/resourceGroups/myRg/providers/Microsoft.Compute/virtualMachines/myVm" -n myRemediation --policy-assignment eeb18edc813c42d0ad5a9eab
  - name: Create a remediation that will re-evaluate compliance before remediating
    text: >
        az policy remediation create -g myRg -n myRemediation --policy-assignment eeb18edc813c42d0ad5a9eab --resource-discovery-mode ReEvaluateCompliance
"""

helps['policy remediation delete'] = """
type: command
short-summary: Delete a resource policy remediation.
"""

helps['policy remediation deployment'] = """
type: group
short-summary: Manage resource policy remediation deployments.
"""

helps['policy remediation deployment list'] = """
type: command
short-summary: Lists deployments for a resource policy remediation.
"""

helps['policy remediation list'] = """
type: command
short-summary: List resource policy remediations.
"""

helps['policy remediation show'] = """
type: command
short-summary: Show a resource policy remediation.
"""

helps['policy state'] = """
type: group
short-summary: Manage policy compliance states.
"""

helps['policy state list'] = """
type: command
short-summary: List policy compliance states.
examples:
  - name: Get latest policy states at current subscription scope.
    text: >
        az policy state list
  - name: Get all policy states at current subscription scope.
    text: >
        az policy state list --all
  - name: Get latest policy states at management group scope.
    text: >
        az policy state list -m "myMg"
  - name: Get latest policy states at resource group scope in current subscription.
    text: >
        az policy state list -g "myRg"
  - name: Get latest policy states for a resource using resource ID.
    text: >
        az policy state list --resource "/subscriptions/fff10b27-fff3-fff5-fff8-fffbe01e86a5/resourceGroups/myResourceGroup /providers/Microsoft.EventHub/namespaces/myns1/eventhubs/eh1/consumergroups/cg1"
  - name: Get latest policy states for a resource using resource name.
    text: >
        az policy state list --resource "myKeyVault" --namespace "Microsoft.KeyVault" --resource-type "vaults" -g "myresourcegroup"
  - name: Get latest policy states for a nested resource using resource name.
    text: >
        az policy state list --resource "myRule1" --namespace "Microsoft.Network" --resource-type "securityRules" --parent "networkSecurityGroups/mysecuritygroup1" -g "myresourcegroup"
  - name: Get latest policy states for a policy set definition in current subscription.
    text: >
        az policy state list -s "fff58873-fff8-fff5-fffc-fffbe7c9d697"
  - name: Get latest policy states for a policy definition in current subscription.
    text: >
        az policy state list -d "fff69973-fff8-fff5-fffc-fffbe7c9d698"
  - name: Get latest policy states for a policy assignment in current subscription.
    text: >
        az policy state list -a "ddd8ef92e3714a5ea3d208c1"
  - name: Get latest policy states for a policy assignment in the specified resource group in current subscription.
    text: >
        az policy state list -g "myRg" -a "ddd8ef92e3714a5ea3d208c1"
  - name: Get top 5 latest policy states in current subscription, selecting a subset of properties and customizing ordering.
    text: >
        az policy state list --top 5 --order-by "timestamp desc, policyAssignmentName asc" --select "timestamp, resourceId, policyAssignmentId, policySetDefinitionId, policyDefinitionId"
  - name: Get latest policy states in current subscription during a custom time interval.
    text: >
        az policy state list --from "2018-03-08T00:00:00Z" --to "2018-03-15T00:00:00Z"
  - name: Get latest policy states in current subscription filtering results based on some property values.
    text: >
        az policy state list --filter "(policyDefinitionAction eq 'deny' or policyDefinitionAction eq 'audit') and resourceLocation ne 'eastus'"
  - name: Get number of latest policy states in current subscription.
    text: >
        az policy state list --apply "aggregate($count as numberOfRecords)"
  - name: Get latest policy states in current subscription aggregating results based on some properties.
    text: >
        az policy state list --apply "groupby((policyAssignmentId, policySetDefinitionId, policyDefinitionReferenceId, policyDefinitionId), aggregate($count as numStates))"
  - name: Get latest policy states in current subscription grouping results based on some properties.
    text: >
        az policy state list --apply "groupby((policyAssignmentName, resourceId))"
  - name: Get latest policy states in current subscription aggregating results based on some properties specifying multiple groupings.
    text: >
        az policy state list --apply "groupby((policyAssignmentId, policySetDefinitionId, policyDefinitionReferenceId, policyDefinitionId, resourceId))/groupby((policyAssignmentId, policySetDefinitionId, policyDefinitionReferenceId, policyDefinitionId), aggregate($count as numNonCompliantResources))"
  - name: Get latest policy states for a resource including policy evaluation details.
    text: >
        az policy state list --resource "myKeyVault" --namespace "Microsoft.KeyVault" --resource-type "vaults" -g "myresourcegroup" --expand PolicyEvaluationDetails
"""

helps['policy state summarize'] = """
type: command
short-summary: Summarize policy compliance states.
examples:
  - name: Get latest non-compliant policy states summary at current subscription scope.
    text: >
        az policy state summarize
  - name: Get latest non-compliant policy states summary at management group scope.
    text: >
        az policy state summarize -m "myMg"
  - name: Get latest non-compliant policy states summary at resource group scope in current subscription.
    text: >
        az policy state summarize -g "myRg"
  - name: Get latest non-compliant policy states summary for a resource using resource ID.
    text: >
        az policy state summarize --resource "/subscriptions/fff10b27-fff3-fff5-fff8-fffbe01e86a5/resourceGroups/myResourceGroup /providers/Microsoft.EventHub/namespaces/myns1/eventhubs/eh1/consumergroups/cg1"
  - name: Get latest non-compliant policy states summary for a resource using resource name.
    text: >
        az policy state summarize --resource "myKeyVault" --namespace "Microsoft.KeyVault" --resource-type "vaults" -g "myresourcegroup"
  - name: Get latest non-compliant policy states summary for a nested resource using resource name.
    text: >
        az policy state summarize --resource "myRule1" --namespace "Microsoft.Network" --resource-type "securityRules" --parent "networkSecurityGroups/mysecuritygroup1" -g "myresourcegroup"
  - name: Get latest non-compliant policy states summary for a policy set definition in current subscription.
    text: >
        az policy state summarize -s "fff58873-fff8-fff5-fffc-fffbe7c9d697"
  - name: Get latest non-compliant policy states summary for a policy definition in current subscription.
    text: >
        az policy state summarize -d "fff69973-fff8-fff5-fffc-fffbe7c9d698"
  - name: Get latest non-compliant policy states summary for a policy assignment in current subscription.
    text: >
        az policy state summarize -a "ddd8ef92e3714a5ea3d208c1"
  - name: Get latest non-compliant policy states summary for a policy assignment in the specified resource group in current subscription.
    text: >
        az policy state summarize -g "myRg" -a "ddd8ef92e3714a5ea3d208c1"
  - name: Get latest non-compliant policy states summary in current subscription, limiting the assignments summary to top 5.
    text: >
        az policy state summarize --top 5
  - name: Get latest non-compliant policy states summary in current subscription for a custom time interval.
    text: >
        az policy state summarize --from "2018-03-08T00:00:00Z" --to "2018-03-15T00:00:00Z"
  - name: Get latest non-compliant policy states summary in current subscription filtering results based on some property values.
    text: >
        az policy state summarize --filter "(policyDefinitionAction eq 'deny' or policyDefinitionAction eq 'audit') and resourceLocation ne 'eastus'"
"""

helps['policy state trigger-scan'] = """
type: command
short-summary: Trigger a policy compliance evaluation for a scope.
examples:
  - name: Trigger a policy compliance evaluation at the current subscription scope.
    text: >
        az policy state trigger-scan
  - name: Trigger a policy compliance evaluation for a resource group.
    text: >
        az policy state trigger-scan -g "myRg"
  - name: Trigger a policy compliance evaluation for a resource group and do not wait for it to complete.
    text: >
        az policy state trigger-scan -g "myRg" --no-wait
"""

helps['policy metadata'] = """
type: group
short-summary: Get policy metadata resources.
"""

helps['policy metadata list'] = """
type: command
short-summary: List policy metadata resources.
examples:
  - name: Get all policy metadata resources.
    text: >
        az policy metadata list
  - name: Get policy metadata resources, limit the output to 5 resources.
    text: >
        az policy metadata list --top 5
"""

helps['policy metadata show'] = """
type: command
short-summary: Get a single policy metadata resource.
examples:
  - name: Get the policy metadata resource with the name 'ACF1000'.
    text: >
        az policy metadata show --name ACF1000
"""



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


AUTOSCALE_TIMEZONES = ["Dateline Standard Time",
                       "UTC-11",
                       "Hawaiian Standard Time",
                       "Alaskan Standard Time",
                       "Pacific Standard Time (Mexico)",
                       "Pacific Standard Time",
                       "US Mountain Standard Time",
                       "Mountain Standard Time (Mexico)",
                       "Mountain Standard Time",
                       "Central America Standard Time",
                       "Central Standard Time",
                       "Central Standard Time (Mexico)",
                       "Canada Central Standard Time",
                       "SA Pacific Standard Time",
                       "Eastern Standard Time",
                       "USEasternStandardTime",
                       "Venezuela Standard Time",
                       "Paraguay Standard Time",
                       "Atlantic Standard Time",
                       "Central Brazilian Standard Time",
                       "SA Western Standard Time",
                       "Pacific SA Standard Time",
                       "Newfoundland Standard Time",
                       "E. South America Standard Time",
                       "Argentina Standard Time",
                       "SA Eastern Standard Time",
                       "Greenland Standard Time",
                       "Montevideo Standard Time",
                       "Bahia Standard Time",
                       "UTC-02",
                       "Mid-Atlantic Standard Time",
                       "Azores Standard Time",
                       "Cape Verde Standard Time",
                       "Morocco Standard Time",
                       "UTC",
                       "GMT Standard Time",
                       "Greenwich Standard Time",
                       "W. Europe Standard Time",
                       "Central Europe Standard Time",
                       "Romance Standard Time",
                       "Central European Standard Time",
                       "W. Central Africa Standard Time",
                       "Namibia Standard Time",
                       "Jordan Standard Time",
                       "GTB Standard Time",
                       "Middle East Standard Time",
                       "Egypt Standard Time",
                       "Syria Standard Time",
                       "E. Europe Standard Time",
                       "South Africa Standard Time",
                       "FLE Standard Time",
                       "Israel Standard Time",
                       "Kaliningrad Standard Time",
                       "Libya Standard Time",
                       "Arabic Standard Time",
                       "Turkey Standard Time",
                       "Arab Standard Time",
                       "Belarus Standard Time",
                       "Russian Standard Time",
                       "E. Africa Standard Time",
                       "Iran Standard Time",
                       "Arabian Standard Time",
                       "Azerbaijan Standard Time",
                       "Russia Time Zone 3",
                       "Mauritius Standard Time",
                       "Georgian Standard Time",
                       "Caucasus Standard Time",
                       "Afghanistan Standard Time",
                       "West Asia Standard Time",
                       "Ekaterinburg Standard Time",
                       "Pakistan Standard Time",
                       "India Standard Time",
                       "Sri Lanka Standard Time",
                       "Nepal Standard Time",
                       "Central Asia Standard Time",
                       "Bangladesh Standard Time",
                       "N. Central Asia Standard Time",
                       "Myanmar Standard Time",
                       "SE Asia Standard Time",
                       "North Asia Standard Time",
                       "China Standard Time",
                       "North Asia East Standard Time",
                       "Singapore Standard Time",
                       "W. Australia Standard Time",
                       "Taipei Standard Time",
                       "Ulaanbaatar Standard Time",
                       "Tokyo Standard Time",
                       "Korea Standard Time",
                       "Yakutsk Standard Time",
                       "Cen. Australia Standard Time",
                       "AUS Central Standard Time",
                       "E. Australia Standard Time",
                       "AUS Eastern Standard Time",
                       "West Pacific Standard Time",
                       "Tasmania Standard Time",
                       "Vladivostok Standard Time",
                       "Russia Time Zone 10",
                       "Magadan Standard Time",
                       "Central Pacific Standard Time",
                       "Russia Time Zone 11",
                       "New Zealand Standard Time",
                       "UTC+12",
                       "Fiji Standard Time",
                       "Kamchatka Standard Time",
                       "Tonga Standard Time",
                       "Samoa Standard Time",
                       "Line Islands Standard Time"
                       ]


def get_key_for_storage_account(cmd, storage_account):  # pylint: disable=unused-argument
    from ._client_factory import cf_storage
    from msrestazure.tools import parse_resource_id, is_valid_resource_id
    from knack.util import CLIError

    storage_account_key = None
    if is_valid_resource_id(storage_account):
        parsed_storage_account = parse_resource_id(storage_account)
        resource_group_name = parsed_storage_account['resource_group']
        storage_account_name = parsed_storage_account['resource_name']

        storage_client = cf_storage(cmd.cli_ctx)
        keys = storage_client.storage_accounts.list_keys(resource_group_name, storage_account_name)
        storage_account_key = keys.keys[0].value  # pylint: disable=no-member
    elif storage_account:
        raise CLIError('Failed to get access key for storage account: {}'.format(storage_account))
    return storage_account_key


def get_storage_account_endpoint(cmd, storage_account, is_wasb):
    from ._client_factory import cf_storage
    from msrestazure.tools import parse_resource_id, is_valid_resource_id
    host = None
    if is_valid_resource_id(storage_account):
        parsed_storage_account = parse_resource_id(storage_account)
        resource_group_name = parsed_storage_account['resource_group']
        storage_account_name = parsed_storage_account['resource_name']

        storage_client = cf_storage(cmd.cli_ctx)
        storage_account = storage_client.storage_accounts.get_properties(
            resource_group_name=resource_group_name,
            account_name=storage_account_name)

        def extract_endpoint(storage_account, is_wasb):
            if not storage_account:
                return None
            return storage_account.primary_endpoints.dfs if not is_wasb else storage_account.primary_endpoints.blob

        def extract_host(uri):
            import re
            return uri and re.search('//(.*)/', uri).groups()[0]

        host = extract_host(extract_endpoint(storage_account, is_wasb))
    return host


def build_identities_info(identities):
    from azure.mgmt.hdinsight.models import ClusterIdentity, ResourceIdentityType
    identity = None
    if identities:
        identity_type = ResourceIdentityType.user_assigned
        identity = ClusterIdentity(type=identity_type)
        identity.user_assigned_identities = {e: {} for e in identities}

    return identity


def build_virtual_network_profile(subnet):
    from msrestazure.tools import resource_id, parse_resource_id, is_valid_resource_id
    from azure.mgmt.hdinsight.models import VirtualNetworkProfile
    from knack.util import CLIError

    vnet_profile = None
    if is_valid_resource_id(subnet):
        parsed_subnet_id = parse_resource_id(subnet)
        subscription_name = parsed_subnet_id['subscription']
        resource_group_name = parsed_subnet_id['resource_group']
        vnet_namespace = parsed_subnet_id['namespace']
        vnet_type = parsed_subnet_id['type']
        vnet_name = parsed_subnet_id['name']
        vnet_id = resource_id(
            subscription=subscription_name,
            resource_group=resource_group_name,
            namespace=vnet_namespace,
            type=vnet_type,
            name=vnet_name)
        vnet_profile = VirtualNetworkProfile(id=vnet_id, subnet=subnet)
    elif subnet:
        raise CLIError('Invalid subnet: {}'.format(subnet))
    return vnet_profile


def parse_domain_name(domain):
    from msrestazure.tools import parse_resource_id, is_valid_resource_id
    domain_name = None
    if is_valid_resource_id(domain):
        parsed_domain_id = parse_resource_id(domain)
        domain_name = parsed_domain_id['resource_name']
    return domain_name


# Validate ESP cluster creation required parameters
def validate_esp_cluster_create_params(esp,
                                       cluster_name,
                                       resource_group_name,
                                       cluster_type,
                                       subnet,
                                       domain,
                                       cluster_admin_account,
                                       assign_identity,
                                       ldaps_urls,
                                       cluster_admin_password,
                                       cluster_users_group_dns):
    from knack.util import CLIError
    if esp:
        missing_params = []
        if not cluster_name:
            missing_params.append("--name/-n")
        if not resource_group_name:
            missing_params.append("--resource-group/-g")
        if not cluster_type:
            missing_params.append("--type/-t")
        if not subnet:
            missing_params.append("--subnet")
        if not domain:
            missing_params.append("--domain")
        if not cluster_admin_account:
            missing_params.append("--cluster-admin-account")
        if not cluster_users_group_dns:
            missing_params.append("--cluster-users-group-dns")
        if not assign_identity:
            missing_params.append("--assign-identity")

        if missing_params:
            raise CLIError('the following params are required  '
                           'when --esp is specified: {}'.format(', '.join(missing_params)))
    else:
        esp_params = []
        if domain:
            esp_params.append("--domain")
        if cluster_admin_account:
            esp_params.append("--cluster-admin_account")
        if ldaps_urls:
            esp_params.append("--ldaps-urls")
        if cluster_admin_password:
            esp_params.append("--cluster-admin-password")
        if cluster_users_group_dns:
            esp_params.append("--cluster-users-group-dns")

        if esp_params:
            raise CLIError('the following params are required only '
                           'when --esp is specified: {}'.format(', '.join(esp_params)))


def get_resource_id_by_name(cli_ctx, resource_type, resource_name):
    from ._client_factory import cf_resources
    from knack.util import CLIError

    client = cf_resources(cli_ctx)
    filter_str = "resourceType eq '{}' and name eq '{}'".format(resource_type, resource_name) if resource_type else None
    resources = list(client.resources.list(filter=filter_str))
    if not resources:
        raise CLIError('Fails to retrieve any resource by name {}'.format(resource_name))
    if len(resources) > 1:
        raise CLIError('Found more than one resources by name {}. '
                       'Please specify one of the following resource IDs explicitly:\n{}'
                       .format(resource_name, '\n'.join([resource.id for resource in resources])))
    return resources[0].id



# coding=utf-8
# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.help_files import helps  # pylint: disable=unused-import
# pylint: disable=line-too-long, too-many-lines

helps['lab'] = """
type: group
short-summary: Manage Azure DevTest Labs.
"""

helps['lab arm-template'] = """
type: group
short-summary: Manage Azure Resource Manager (ARM) templates in an Azure DevTest Lab.
"""

helps['lab arm-template show'] = """
type: command
short-summary: Get the details of an ARM template in a lab.
parameters:
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --name -n
    short-summary: Name of the Azure Resource Manager template.
  - name: --resource-group -g
    short-summary: Name of lab's resource group.
  - name: --export-parameters
    short-summary: Whether or not to export parameters template.
  - name: --artifact-source-name
    short-summary: Name of the artifact source.
"""

helps['lab artifact'] = """
type: group
short-summary: Manage DevTest Labs artifacts.
"""

helps['lab artifact-source'] = """
type: group
short-summary: Manage DevTest Lab artifact sources.
"""

helps['lab custom-image'] = """
type: group
short-summary: Manage custom images of a DevTest Lab.
"""

helps['lab custom-image create'] = """
type: command
short-summary: Create a custom image in a DevTest Lab.
parameters:
  - name: --name -n
    short-summary: Name of the image.
  - name: --lab-name
    short-summary: Name of the Lab.
  - name: --author
    short-summary: The author of the custom image.
  - name: --description
    short-summary: A detailed description for the custom image.
  - name: --source-vm-id
    short-summary: The resource ID of a virtual machine in the provided lab.
  - name: --os-type
    short-summary: 'Type of the OS on which the custom image is based. Allowed values are: Windows, Linux'
  - name: --os-state
    short-summary: The current state of the virtual machine.
    long-summary: >
        For Windows virtual machines: NonSysprepped, SysprepRequested, SysprepApplied
        For Linux virtual machines: NonDeprovisioned, DeprovisionRequested, DeprovisionApplied
examples:
  - name: Create a custom image in the lab from a running Windows virtual machine without applying sysprep.
    text: |
        az lab custom-image create --lab-name {LabName} -g {ResourceGroup} --name {VMName} \\
            --os-type Windows --os-state NonSysprepped \\
            --source-vm-id "/subscriptions/{SubID}/resourcegroups/{ResourceGroup}/microsoft.devtestlab/labs/{LabName}/virtualmachines/{VMName}"
"""

helps['lab environment'] = """
type: group
short-summary: Manage environments for an Azure DevTest Lab.
"""

helps['lab environment create'] = """
type: command
short-summary: Create an environment in a lab.
parameters:
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --name -n
    short-summary: Name of the environment.
  - name: --resource-group -g
    short-summary: Name of the lab's resource group.
  - name: --arm-template
    short-summary: Name or ID of the ARM template in the lab.
  - name: --artifact-source-name
    short-summary: Name of the artifact source in the lab.
    populator-commands:
      - az lab artifact-source list
  - name: --parameters
    short-summary: JSON encoded list of parameters. Use '@{file}' to load from a file.
  - name: --tags
    short-summary: The tags for the resource.
"""

helps['lab environment delete'] = """
type: command
short-summary: Delete an environment from a lab.
"""

helps['lab environment list'] = """
type: command
short-summary: List environments in a lab.
examples:
  - name: List environments in a lab. (autogenerated)
    text: az lab environment list --lab-name MyLab --resource-group MyResourceGroup
    crafted: true
"""

helps['lab environment show'] = """
type: command
short-summary: Get the details for an environment of a lab.
"""

helps['lab formula'] = """
type: group
short-summary: Manage formulas for an Azure DevTest Lab.
"""

helps['lab formula export-artifacts'] = """
type: command
short-summary: Export artifacts from a formula.
parameters:
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --name -n
    short-summary: Name of the formula.
examples:
  - name: Export artifacts from a formula. (autogenerated)
    text: az lab formula export-artifacts --lab-name MyLab --name MyFormula --resource-group MyResourceGroup
    crafted: true
"""

helps['lab formula show'] = """
type: command
short-summary: Show formulae from an Azure DevTest Lab.
parameters:
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --name -n
    short-summary: Name of the formula.
examples:
  - name: Show formulae from an Azure DevTest Lab. (autogenerated)
    text: az lab formula show --lab-name MyLab --name MyFormula --resource-group MyResourceGroup
    crafted: true
"""

helps['lab gallery-image'] = """
type: group
short-summary: List Azure Marketplace images allowed for a DevTest Lab.
"""

helps['lab secret'] = """
type: group
short-summary: Manage secrets of an Azure DevTest Lab.
"""

helps['lab secret set'] = """
type: command
short-summary: Set a secret for a lab.
parameters:
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --name -n
    short-summary: Name of the secret.
  - name: --value
    short-summary: Value of the secret.
"""

helps['lab vm'] = """
type: group
short-summary: Manage VMs in an Azure DevTest Lab.
"""

helps['lab vm apply-artifacts'] = """
type: command
short-summary: Apply artifacts to a virtual machine in Azure DevTest Lab.
parameters:
  - name: --resource-group -g
    short-summary: Name of lab's resource group.
  - name: --lab-name
    short-summary: Name of the Lab.
  - name: --name -n
    short-summary: Name of the virtual machine.
  - name: --artifacts
    short-summary: JSON encoded array of artifacts to be applied. Use '@{file}' to load from a file.
examples:
  - name: Apply artifacts to a virtual machine in Azure DevTest Lab. (autogenerated)
    text: az lab vm apply-artifacts --artifacts '@artifacts.json' --lab-name MyLab --name MyVirtualMachine --resource-group MyResourceGroup
    crafted: true
"""

helps['lab vm claim'] = """
type: command
short-summary: Claim a virtual machine from the Lab.
parameters:
  - name: --resource-group -g
    short-summary: Name of lab's resource group.
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --name -n
    short-summary: Name of the virtual machine to claim.
examples:
  - name: Claim any available virtual machine in the lab.
    text: >
        az lab vm claim -g {ResourceGroup} --lab-name {LabName}
  - name: Claim a specific virtual machine in the lab.
    text: >
        az lab vm claim -g {ResourceGroup} --lab-name {LabName} --name {VMName}
  - name: Claim multiple virtual machines in the lab by IDs.
    text: |
        az lab vm claim --ids \\
            /subscriptions/{SubID}/resourcegroups/{ResourceGroup}/providers/microsoft.devtestlab/labs/{LabName}/virtualmachines/{VMName1} \\
            /subscriptions/{SubID}/resourcegroups/{ResourceGroup}/providers/microsoft.devtestlab/labs/{LabName}/virtualmachines/{VMName2}
"""

helps['lab vm create'] = """
type: command
short-summary: Create a VM in a lab.
parameters:
  - name: --name -n
    short-summary: Name of the virtual machine.
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --notes
    short-summary: Notes for the virtual machine.
  - name: --image
    short-summary: The name of the operating system image (gallery image name or custom image name/ID).
    long-summary: Use `az lab gallery-image list` for available gallery images or `az lab custom-image list` for available custom images.
  - name: --image-type
    short-summary: 'Type of the image. Allowed values are: gallery, custom'
  - name: --formula
    short-summary: Name of the formula. Use `az lab formula list` for available formulas.
    long-summary: >
        Use `az lab formula` with the `--export-artifacts` flag to export and update artifacts, then pass
        the results via the `--artifacts` argument.
  - name: --size
    short-summary: 'The size of the VM to be created. See https://azure.microsoft.com/pricing/details/virtual-machines/ for size info.'
  - name: --admin-username
    short-summary: Username for the VM admin.
  - name: --admin-password
    short-summary: Password for the VM admin.
  - name: --ssh-key
    short-summary: The SSH public key or public key file path. Use `--generate-ssh-keys` to generate SSH keys.
  - name: --authentication-type
    short-summary: 'Type of authentication allowed for the VM. Allowed values are: password, ssh.'
  - name: --saved-secret
    short-summary: Name of the saved secret to be used for authentication.
    long-summary: When this value is provided, it is used in the place of other authentication methods.
  - name: --vnet-name
    short-summary: Name of the virtual network to add the VM to.
  - name: --subnet
    short-summary: Name of the subnet to add the VM to.
  - name: --ip-configuration
    short-summary: 'Type of IP configuration to use for the VM. Allowed values are: shared, public, private.'
    long-summary: If omitted, will be selected based on the VM's vnet.
  - name: --artifacts
    short-summary: JSON encoded array of artifacts to be applied. Use '@{file}' to load from a file.
  - name: --tags
    short-summary: Space-separated tags in `key[=value]` format.
    long-summary: Tags may be cleared by assigning the empty value "" to them.
  - name: --allow-claim
    short-summary: Flag indicating if the VM should be created as claimable.
  - name: --disk-type
    short-summary: Storage type to use for virtual machine.
  - name: --expiration-date
    short-summary: The expiration date in UTC(yyyy-MM-ddTHH:mm:ss) for the VM.
  - name: --generate-ssh-keys
    short-summary: Generate SSH public and private key files if missing.
examples:
  - name: Create a VM in the lab from a gallery image.
    text: >
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 16.04 LTS" --image-type gallery --size Standard_DS1_v2
  - name: Create a VM in the lab from a gallery image with SSH authentication.
    text: >
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 16.04 LTS" --image-type gallery --size Standard_DS1_v2 --authentication-type ssh
  - name: Create a claimable VM in the lab from a gallery image with password authentication.
    text: >
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 16.04 LTS" --image-type gallery --size Standard_DS1_v2 --allow-claim
  - name: Create a windows VM in the lab from a gallery image with password authentication.
    text: >
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Windows Server 2008 R2 SP1" --image-type gallery --size Standard_DS1_v2
  - name: Create a VM in the lab from a custom image.
    text: >
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "jenkins_custom" --image-type custom --size Standard_DS1_v2
  - name: Create a VM in the lab with a public IP.
    text: >
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 16.04 LTS" --image-type gallery --size Standard_DS1_v2 --ip-configuration public
  - name: Create a VM from a formula.
    text: >
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --formula MyFormula --artifacts '@artifacts.json'
"""

helps['lab vm list'] = """
type: command
short-summary: List the VMs in an Azure DevTest Lab.
parameters:
  - name: --lab-name
    short-summary: Name of the lab.
  - name: --order-by
    short-summary: The ordering expression for the results using OData notation.
  - name: --top
    short-summary: The maximum number of resources to return.
  - name: --filters
    short-summary: The filter to apply.
  - name: --expand
    short-summary: The expand query.
  - name: --claimable
    short-summary: List only claimable virtual machines in the lab. Cannot be used with `--filters`.
  - name: --all
    short-summary: List all virtual machines in the lab. Cannot be used with `--filters`
  - name: --environment
    short-summary: Name or ID of the environment to list virtual machines in. Cannot be used with `--filters`.
  - name: --object-id
    short-summary: Object ID of the owner to list VMs for.
examples:
  - name: List the VMs in an Azure DevTest Lab. (autogenerated)
    text: az lab vm list --lab-name MyLab --resource-group MyResourceGroup
    crafted: true
"""

helps['lab vnet'] = """
type: group
short-summary: Manage virtual networks of an Azure DevTest Lab.
"""



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from ..azcopy.util import AzCopy, client_auth_for_azcopy, login_auth_for_azcopy, _generate_sas_token


# pylint: disable=too-many-statements, too-many-locals, unused-argument
def storage_copy(source, destination, put_md5=None, recursive=None, blob_type=None,
                 preserve_s2s_access_tier=None, content_type=None, follow_symlinks=None,
                 exclude_pattern=None, include_pattern=None, exclude_path=None, include_path=None, **kwargs):

    azcopy = AzCopy()
    flags = []
    if recursive is not None:
        flags.append('--recursive')
    if put_md5 is not None:
        flags.append('--put-md5')
    if blob_type is not None:
        flags.append('--blob-type=' + blob_type)
    if preserve_s2s_access_tier is not None:
        flags.append('--s2s-preserve-access-tier=' + str(preserve_s2s_access_tier))
    if include_pattern is not None:
        flags.append('--include-pattern=' + include_pattern)
    if exclude_pattern is not None:
        flags.append('--exclude-pattern=' + exclude_pattern)
    if include_path is not None:
        flags.append('--include-path=' + include_path)
    if exclude_pattern is not None:
        flags.append('--exclude-path=' + exclude_path)
    if content_type is not None:
        flags.append('--content-type=' + content_type)
    if follow_symlinks is not None:
        flags.append('--follow-symlinks=true')
    azcopy.copy(source, destination, flags=flags)


def storage_remove(cmd, client, service, target, recursive=None, exclude_pattern=None, include_pattern=None,
                   exclude_path=None, include_path=None):
    if service == 'file':
        azcopy = _azcopy_file_client(cmd, client)
    else:
        azcopy = _azcopy_blob_client(cmd, client)
    flags = []
    if recursive is not None:
        flags.append('--recursive')
    if include_pattern is not None:
        flags.append('--include-pattern=' + include_pattern)
    if exclude_pattern is not None:
        flags.append('--exclude-pattern=' + exclude_pattern)
    if include_path is not None:
        flags.append('--include-path=' + include_path)
    if exclude_path is not None:
        flags.append('--exclude-path=' + exclude_path)

    sas_token = client.sas_token

    if not sas_token and client.account_key:
        sas_token = _generate_sas_token(cmd, client.account_name, client.account_key, service=service,
                                        resource_types='co',
                                        permissions='rdl')

    azcopy.remove(_add_url_sas(target, sas_token), flags=flags)


# pylint: disable=unused-argument
def storage_fs_directory_copy(cmd, source, destination, recursive=None, **kwargs):
    azcopy = AzCopy()
    if kwargs.get('token_credential'):
        azcopy = _azcopy_login_client(cmd)

    flags = []
    if recursive:
        flags.append('--recursive')
    azcopy.copy(source, destination, flags=flags)


def storage_blob_sync(cmd, client, source, destination, exclude_pattern=None, include_pattern=None,
                      exclude_path=None):
    azcopy = _azcopy_blob_client(cmd, client)
    flags = ['--delete-destination=true']
    if include_pattern is not None:
        flags.append('--include-pattern=' + include_pattern)
    if exclude_pattern is not None:
        flags.append('--exclude-pattern=' + exclude_pattern)
    if exclude_path is not None:
        flags.append('--exclude-path=' + exclude_path)

    sas_token = client.sas_token

    if not sas_token and client.account_key:
        sas_token = _generate_sas_token(cmd, client.account_name, client.account_key, service='blob',
                                        resource_types='co',
                                        permissions='rwdlac')

    azcopy.sync(source, _add_url_sas(destination, sas_token), flags=flags)


def storage_run_command(cmd, command_args):
    if command_args.startswith('azcopy'):
        command_args = command_args[len('azcopy'):]
    azcopy = _azcopy_login_client(cmd)
    azcopy.run_command([command_args])


def _add_url_sas(url, sas):
    if not sas:
        return url
    return '{}?{}'.format(url, sas)


def _azcopy_blob_client(cmd, client):
    return AzCopy(creds=client_auth_for_azcopy(cmd, client))


def _azcopy_file_client(cmd, client):
    return AzCopy(creds=client_auth_for_azcopy(cmd, client, service='file'))


def _azcopy_login_client(cmd):
    return AzCopy(creds=login_auth_for_azcopy(cmd))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""Table transformer for storage commands"""

from azure.cli.core.profiles import get_sdk, ResourceType
from knack.log import get_logger

logger = get_logger(__name__)


def build_table_output(result, projection):

    if not isinstance(result, list):
        result = [result]

    final_list = []

    from collections import OrderedDict
    for item in result:
        def _value_from_path(each_item, path):
            obj = each_item
            try:
                for part in path.split('.'):
                    obj = obj.get(part, None)
            except AttributeError:
                obj = None
            return obj or ' '

        item_dict = OrderedDict()
        for element in projection:
            item_dict[element[0]] = _value_from_path(item, element[1])
        final_list.append(item_dict)

    return final_list


def transform_container_list(result):
    return build_table_output(result, [
        ('Name', 'name'),
        ('Lease Status', 'properties.leaseStatus'),
        ('Last Modified', 'properties.lastModified')
    ])


def transform_container_show(result):
    return build_table_output(result, [
        ('Name', 'name'),
        ('Lease Status', 'properties.lease.status'),
        ('Last Modified', 'properties.lastModified')
    ])


def transform_blob_output(result):
    return build_table_output(result, [
        ('Name', 'name'),
        ('Blob Type', 'properties.blobType'),
        ('Blob Tier', 'properties.blobTier'),
        ('Length', 'properties.contentLength'),
        ('Content Type', 'properties.contentSettings.contentType'),
        ('Last Modified', 'properties.lastModified'),
        ('Snapshot', 'snapshot')
    ])


def transform_share_list(result):
    return build_table_output(result, [
        ('Name', 'name'),
        ('Quota', 'properties.quota'),
        ('Last Modified', 'properties.lastModified')
    ])


def transform_file_output(result):
    """ Transform to convert SDK file/dir list output to something that
    more clearly distinguishes between files and directories. """
    from collections import OrderedDict
    new_result = []

    iterable = result if isinstance(result, list) else result.get('items', result)
    for item in iterable:
        new_entry = OrderedDict()

        entity_type = item['type']  # type property is added by transform_file_directory_result
        is_dir = entity_type == 'dir'

        new_entry['Name'] = item['name'] + '/' if is_dir else item['name']
        new_entry['Content Length'] = ' ' if is_dir else item['properties']['contentLength']
        new_entry['Type'] = item['type']
        new_entry['Last Modified'] = item['properties']['lastModified'] or ' '
        new_result.append(new_entry)
    return sorted(new_result, key=lambda k: k['Name'])


def transform_entity_show(result):
    from collections import OrderedDict
    timestamp = result.pop('Timestamp')
    result.pop('etag')

    # Reassemble the output
    new_result = OrderedDict()
    new_result['Partition'] = result.pop('PartitionKey')
    new_result['Row'] = result.pop('RowKey')
    for key in sorted(result.keys()):
        new_result[key] = result[key]
    new_result['Timestamp'] = timestamp
    return new_result


def transform_message_show(result):
    from collections import OrderedDict
    ordered_result = []
    for item in result:
        new_result = OrderedDict()
        new_result['MessageId'] = item.pop('id')
        new_result['Content'] = item.pop('content')
        new_result['InsertionTime'] = item.pop('insertionTime')
        new_result['ExpirationTime'] = item.pop('expirationTime')
        for key in sorted(item.keys()):
            new_result[key] = item[key]
        ordered_result.append(new_result)
    return ordered_result


def transform_boolean_for_table(result):
    for key in result:
        result[key] = str(result[key])
    return result


def transform_file_directory_result(cli_ctx):
    """
    Transform a the result returned from file and directory listing API.

    This transformer add and remove properties from File and Directory objects in the given list
    in order to align the object's properties so as to offer a better view to the file and dir
    list.
    """
    def transformer(result):
        if getattr(result, 'next_marker', None):
            logger.warning('Next Marker:')
            logger.warning(result.next_marker)

        t_file, t_dir = get_sdk(cli_ctx, ResourceType.DATA_STORAGE, 'File', 'Directory', mod='file.models')
        return_list = []
        for each in result:
            if isinstance(each, t_file):
                delattr(each, 'content')
                setattr(each, 'type', 'file')
            elif isinstance(each, t_dir):
                setattr(each, 'type', 'dir')
            return_list.append(each)

        return return_list
    return transformer



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from distutils.version import StrictVersion  # pylint: disable=no-name-in-module,import-error

# pylint: disable=no-name-in-module,import-error
from azure.mgmt.containerservice.v2021_02_01.models import ManagedClusterLoadBalancerProfile
from azure.mgmt.containerservice.v2021_02_01.models import ManagedClusterLoadBalancerProfileManagedOutboundIPs
from azure.mgmt.containerservice.v2021_02_01.models import ManagedClusterLoadBalancerProfileOutboundIPPrefixes
from azure.mgmt.containerservice.v2021_02_01.models import ManagedClusterLoadBalancerProfileOutboundIPs
from azure.mgmt.containerservice.v2021_02_01.models import ResourceReference

from knack.log import get_logger

logger = get_logger(__name__)


def set_load_balancer_sku(sku, kubernetes_version):
    if sku:
        return sku
    if kubernetes_version and StrictVersion(kubernetes_version) < StrictVersion("1.13.0"):
        logger.warning('Setting load_balancer_sku to basic as it is not specified and kubernetes'
                       'version(%s) less than 1.13.0 only supports basic load balancer SKU\n',
                       kubernetes_version)
        return "basic"
    return "standard"


def update_load_balancer_profile(managed_outbound_ip_count, outbound_ips, outbound_ip_prefixes,
                                 outbound_ports, idle_timeout, profile):
    """parse and update an existing load balancer profile"""
    if not is_load_balancer_profile_provided(managed_outbound_ip_count, outbound_ips, outbound_ip_prefixes,
                                             outbound_ports, idle_timeout):
        return profile
    return configure_load_balancer_profile(managed_outbound_ip_count, outbound_ips, outbound_ip_prefixes,
                                           outbound_ports, idle_timeout, profile)


def create_load_balancer_profile(managed_outbound_ip_count, outbound_ips, outbound_ip_prefixes,
                                 outbound_ports, idle_timeout):
    """parse and build load balancer profile"""
    if not is_load_balancer_profile_provided(managed_outbound_ip_count, outbound_ips, outbound_ip_prefixes,
                                             outbound_ports, idle_timeout):
        return None

    profile = ManagedClusterLoadBalancerProfile()
    return configure_load_balancer_profile(managed_outbound_ip_count, outbound_ips, outbound_ip_prefixes,
                                           outbound_ports, idle_timeout, profile)


def configure_load_balancer_profile(managed_outbound_ip_count, outbound_ips, outbound_ip_prefixes, outbound_ports,
                                    idle_timeout, profile):
    """configure a load balancer with customer supplied values"""
    if not profile:
        return profile

    outbound_ip_resources = _get_load_balancer_outbound_ips(outbound_ips)
    outbound_ip_prefix_resources = _get_load_balancer_outbound_ip_prefixes(outbound_ip_prefixes)

    if managed_outbound_ip_count or outbound_ip_resources or outbound_ip_prefix_resources:
        profile.managed_outbound_ips = None
        profile.outbound_ips = None
        profile.outbound_ip_prefixes = None
        if managed_outbound_ip_count:
            profile.managed_outbound_ips = ManagedClusterLoadBalancerProfileManagedOutboundIPs(
                count=managed_outbound_ip_count
            )
        if outbound_ip_resources:
            profile.outbound_ips = ManagedClusterLoadBalancerProfileOutboundIPs(
                public_ips=outbound_ip_resources
            )
        if outbound_ip_prefix_resources:
            profile.outbound_ip_prefixes = ManagedClusterLoadBalancerProfileOutboundIPPrefixes(
                public_ip_prefixes=outbound_ip_prefix_resources
            )
    if outbound_ports:
        profile.allocated_outbound_ports = outbound_ports
    if idle_timeout:
        profile.idle_timeout_in_minutes = idle_timeout
    return profile


def is_load_balancer_profile_provided(managed_outbound_ip_count, outbound_ips, ip_prefixes,
                                      outbound_ports, idle_timeout):
    return any([managed_outbound_ip_count,
                outbound_ips,
                ip_prefixes,
                outbound_ports,
                idle_timeout])


def _get_load_balancer_outbound_ips(load_balancer_outbound_ips):
    """parse load balancer profile outbound IP ids and return an array of references to the outbound IP resources"""
    load_balancer_outbound_ip_resources = None
    if load_balancer_outbound_ips:
        load_balancer_outbound_ip_resources = \
            [ResourceReference(id=x.strip()) for x in load_balancer_outbound_ips.split(',')]
    return load_balancer_outbound_ip_resources


def _get_load_balancer_outbound_ip_prefixes(load_balancer_outbound_ip_prefixes):
    """parse load balancer profile outbound IP prefix ids and return an array \
    of references to the outbound IP prefix resources"""
    load_balancer_outbound_ip_prefix_resources = None
    if load_balancer_outbound_ip_prefixes:
        load_balancer_outbound_ip_prefix_resources = \
            [ResourceReference(id=x.strip()) for x in load_balancer_outbound_ip_prefixes.split(',')]
    return load_balancer_outbound_ip_prefix_resources



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long
# pylint: disable=too-many-lines
# pylint: disable=inconsistent-return-statements
# pylint: disable=unused-variable

import re


# Namespace Region
def cli_namespace_create(client, resource_group_name, namespace_name, location=None, tags=None, sku='Standard',
                         capacity=None, default_action=None):

    from azure.mgmt.servicebus.models import SBNamespace, SBSku
    client.create_or_update(
        resource_group_name=resource_group_name,
        namespace_name=namespace_name,
        parameters=SBNamespace(
            location=location,
            tags=tags,
            sku=SBSku(
                name=sku,
                tier=sku,
                capacity=capacity)
        )
    ).result()

    if default_action:
        netwrokruleset = client.get_network_rule_set(resource_group_name, namespace_name)
        netwrokruleset.default_action = default_action
        client.create_or_update_network_rule_set(resource_group_name, namespace_name, netwrokruleset)

    return client.get(resource_group_name, namespace_name)


def cli_namespace_update(client, instance, tags=None, sku=None, capacity=None, default_action=None):
    from msrestazure.tools import parse_resource_id

    if tags is not None:
        instance.tags = tags

    if sku is not None:
        instance.sku.name = sku
        instance.sku.tier = sku

    if capacity is not None:
        instance.sku.capacity = capacity

    if default_action:
        resourcegroup = parse_resource_id(instance.id)['resource_group']
        netwrokruleset = client.get_network_rule_set(resourcegroup, instance.name)
        netwrokruleset.default_action = default_action
        client.create_or_update_network_rule_set(resourcegroup, instance.name, netwrokruleset)

    return instance


def cli_namespace_list(client, resource_group_name=None):
    if resource_group_name:
        return client.list_by_resource_group(resource_group_name=resource_group_name)

    return client.list()


# Namespace Authorization rule:
def cli_namespaceautho_create(client, resource_group_name, namespace_name, name, access_rights=None):
    from azure.cli.command_modules.servicebus._utils import accessrights_converter
    return client.create_or_update_authorization_rule(
        resource_group_name=resource_group_name,
        namespace_name=namespace_name,
        authorization_rule_name=name,
        rights=accessrights_converter(access_rights)
    )


# Namespace Authorization rule:
def cli_namespaceautho_update(instance, rights):
    from azure.cli.command_modules.servicebus._utils import accessrights_converter
    instance.rights = accessrights_converter(rights)
    return instance


# Queue Region
def cli_sbqueue_create(cmd, client, resource_group_name, namespace_name, queue_name, lock_duration=None,
                       max_size_in_megabytes=None, requires_duplicate_detection=None, requires_session=None,
                       default_message_time_to_live=None, dead_lettering_on_message_expiration=None,
                       duplicate_detection_history_time_window=None, max_delivery_count=None, status=None,
                       auto_delete_on_idle=None, enable_partitioning=None, enable_express=None,
                       forward_to=None, forward_dead_lettered_messages_to=None, enable_batched_operations=None):

    from azure.mgmt.servicebus.models import SBQueue

    if max_size_in_megabytes:
        cli_returnnsdetails(cmd, resource_group_name, namespace_name, max_size_in_megabytes)

    queue_params = SBQueue(
        lock_duration=return_valid_duration_create(lock_duration),
        max_size_in_megabytes=max_size_in_megabytes,
        requires_duplicate_detection=requires_duplicate_detection,
        requires_session=requires_session,
        default_message_time_to_live=return_valid_duration_create(default_message_time_to_live),
        dead_lettering_on_message_expiration=dead_lettering_on_message_expiration,
        duplicate_detection_history_time_window=return_valid_duration_create(duplicate_detection_history_time_window),
        max_delivery_count=max_delivery_count,
        status=status,
        auto_delete_on_idle=return_valid_duration_create(auto_delete_on_idle),
        enable_partitioning=enable_partitioning,
        enable_express=enable_express,
        forward_to=forward_to,
        forward_dead_lettered_messages_to=forward_dead_lettered_messages_to,
        enable_batched_operations=enable_batched_operations
    )
    return client.create_or_update(
        resource_group_name=resource_group_name,
        namespace_name=namespace_name,
        queue_name=queue_name,
        parameters=queue_params)


def cli_sbqueue_update(instance, lock_duration=None,
                       max_size_in_megabytes=None, requires_duplicate_detection=None, requires_session=None,
                       default_message_time_to_live=None, dead_lettering_on_message_expiration=None,
                       duplicate_detection_history_time_window=None, max_delivery_count=None, status=None,
                       auto_delete_on_idle=None, enable_partitioning=None, enable_express=None,
                       forward_to=None, forward_dead_lettered_messages_to=None, enable_batched_operations=None):

    instance.lock_duration = return_valid_duration(instance.lock_duration, lock_duration)

    if max_size_in_megabytes:
        instance.max_size_in_megabytes = max_size_in_megabytes

    if requires_duplicate_detection:
        instance.requires_duplicate_detection = requires_duplicate_detection

    if requires_session:
        instance.requires_session = requires_session

    instance.default_message_time_to_live = return_valid_duration(instance.default_message_time_to_live, default_message_time_to_live)

    if dead_lettering_on_message_expiration:
        instance.dead_lettering_on_message_expiration = dead_lettering_on_message_expiration

    instance.duplicate_detection_history_time_window = return_valid_duration(instance.duplicate_detection_history_time_window, duplicate_detection_history_time_window)

    if max_delivery_count:
        instance.max_delivery_count = max_delivery_count

    if status:
        instance.status = status

    instance.auto_delete_on_idle = return_valid_duration(instance.auto_delete_on_idle, auto_delete_on_idle)

    if enable_partitioning:
        instance.enable_partitioning = enable_partitioning

    if enable_express:
        instance.enable_express = enable_express

    if forward_to:
        instance.forward_to = forward_to

    if forward_dead_lettered_messages_to:
        instance.forward_dead_lettered_messages_to = forward_dead_lettered_messages_to

    if enable_batched_operations:
        instance.enable_batched_operations = enable_batched_operations

    return instance


# Topic Region
def cli_sbtopic_create(cmd, client, resource_group_name, namespace_name, topic_name, default_message_time_to_live=None,
                       max_size_in_megabytes=None, requires_duplicate_detection=None,
                       duplicate_detection_history_time_window=None,
                       enable_batched_operations=None, status=None, support_ordering=None, auto_delete_on_idle=None,
                       enable_partitioning=None, enable_express=None):
    from azure.mgmt.servicebus.models import SBTopic

    if max_size_in_megabytes:
        cli_returnnsdetails(cmd, resource_group_name, namespace_name, max_size_in_megabytes)

    topic_params = SBTopic(
        default_message_time_to_live=return_valid_duration_create(default_message_time_to_live),
        max_size_in_megabytes=max_size_in_megabytes,
        requires_duplicate_detection=requires_duplicate_detection,
        duplicate_detection_history_time_window=return_valid_duration_create(duplicate_detection_history_time_window),
        enable_batched_operations=enable_batched_operations,
        status=status,
        support_ordering=support_ordering,
        auto_delete_on_idle=return_valid_duration_create(auto_delete_on_idle),
        enable_partitioning=enable_partitioning,
        enable_express=enable_express
    )
    return client.create_or_update(
        resource_group_name=resource_group_name,
        namespace_name=namespace_name,
        topic_name=topic_name,
        parameters=topic_params)


def cli_sbtopic_update(instance, default_message_time_to_live=None,
                       max_size_in_megabytes=None, requires_duplicate_detection=None,
                       duplicate_detection_history_time_window=None,
                       enable_batched_operations=None, status=None, support_ordering=None, auto_delete_on_idle=None,
                       enable_partitioning=None, enable_express=None):

    instance.default_message_time_to_live = return_valid_duration(instance.default_message_time_to_live, default_message_time_to_live)

    if max_size_in_megabytes:
        instance.max_size_in_megabytes = max_size_in_megabytes

    if requires_duplicate_detection:
        instance.requires_duplicate_detection = requires_duplicate_detection

    instance.duplicate_detection_history_time_window = return_valid_duration(instance.duplicate_detection_history_time_window, duplicate_detection_history_time_window)

    if enable_batched_operations:
        instance.enable_batched_operations = enable_batched_operations

    if status:
        instance.status = status

    if support_ordering:
        instance.support_ordering = support_ordering

    instance.auto_delete_on_idle = return_valid_duration(instance.auto_delete_on_idle, auto_delete_on_idle)

    if enable_partitioning:
        instance.enable_partitioning = enable_partitioning

    if enable_express:
        instance.enable_express = enable_express

    return instance


# Subscription Region
def cli_sbsubscription_create(client, resource_group_name, namespace_name, topic_name, subscription_name, lock_duration=None,
                              requires_session=None, default_message_time_to_live=None, dead_lettering_on_message_expiration=None,
                              max_delivery_count=None, status=None, enable_batched_operations=None,
                              auto_delete_on_idle=None, forward_to=None, forward_dead_lettered_messages_to=None, dead_lettering_on_filter_evaluation_exceptions=None):

    from azure.mgmt.servicebus.models import SBSubscription
    subscription_params = SBSubscription(
        lock_duration=return_valid_duration_create(lock_duration),
        requires_session=requires_session,
        default_message_time_to_live=return_valid_duration_create(default_message_time_to_live),
        dead_lettering_on_message_expiration=dead_lettering_on_message_expiration,
        max_delivery_count=max_delivery_count,
        status=status,
        enable_batched_operations=enable_batched_operations,
        auto_delete_on_idle=return_valid_duration_create(auto_delete_on_idle),
        forward_to=forward_to,
        forward_dead_lettered_messages_to=forward_dead_lettered_messages_to,
        dead_lettering_on_filter_evaluation_exceptions=dead_lettering_on_filter_evaluation_exceptions
    )

    return client.create_or_update(
        resource_group_name=resource_group_name,
        namespace_name=namespace_name,
        topic_name=topic_name,
        subscription_name=subscription_name,
        parameters=subscription_params)


def cli_sbsubscription_update(instance, lock_duration=None,
                              requires_session=None, default_message_time_to_live=None,
                              dead_lettering_on_message_expiration=None,
                              max_delivery_count=None, status=None, enable_batched_operations=None,
                              auto_delete_on_idle=None, forward_to=None, forward_dead_lettered_messages_to=None, dead_lettering_on_filter_evaluation_exceptions=None):

    instance.lock_duration = return_valid_duration(instance.lock_duration, lock_duration)

    if requires_session:
        instance.requires_session = requires_session

    instance.default_message_time_to_live = return_valid_duration(instance.default_message_time_to_live, default_message_time_to_live)

    if dead_lettering_on_message_expiration:
        instance.dead_lettering_on_message_expiration = dead_lettering_on_message_expiration

    if max_delivery_count:
        instance.max_delivery_count = max_delivery_count

    if status:
        instance.status = status

    if enable_batched_operations:
        instance.enable_batched_operations = enable_batched_operations

    instance.auto_delete_on_idle = return_valid_duration(instance.auto_delete_on_idle, auto_delete_on_idle)

    if forward_to:
        instance.forward_to = forward_to

    if forward_dead_lettered_messages_to:
        instance.forward_dead_lettered_messages_to = forward_dead_lettered_messages_to

    if dead_lettering_on_filter_evaluation_exceptions is not None:
        instance.dead_lettering_on_filter_evaluation_exceptions = dead_lettering_on_filter_evaluation_exceptions

    return instance


# Rule Region
def cli_rules_create(client, resource_group_name, namespace_name, topic_name, subscription_name, rule_name,
                     action_sql_expression=None, action_compatibility_level=None, action_requires_preprocessing=None,
                     filter_sql_expression=None, filter_requires_preprocessing=None, correlation_id=None,
                     message_id=None, to=None, reply_to=None, label=None, session_id=None, reply_to_session_id=None,
                     content_type=None, requires_preprocessing=None):

    from azure.mgmt.servicebus.models import Rule, Action, SqlFilter, CorrelationFilter
    parameters = Rule()
    parameters.action = Action(
        sql_expression=action_sql_expression,
        compatibility_level=action_compatibility_level,
        requires_preprocessing=action_requires_preprocessing
    )
    parameters.sql_filter = SqlFilter(
        sql_expression=filter_sql_expression,
        requires_preprocessing=filter_requires_preprocessing
    )
    parameters.correlation_filter = CorrelationFilter(
        correlation_id=correlation_id,
        to=to,
        message_id=message_id,
        reply_to=reply_to,
        label=label,
        session_id=session_id,
        reply_to_session_id=reply_to_session_id,
        content_type=content_type,
        requires_preprocessing=requires_preprocessing
    )
    return client.create_or_update(
        resource_group_name=resource_group_name,
        namespace_name=namespace_name,
        topic_name=topic_name,
        subscription_name=subscription_name,
        rule_name=rule_name,
        parameters=parameters)


# Rule Region
def cli_rules_update(instance,
                     action_sql_expression=None, action_compatibility_level=None, action_requires_preprocessing=None,
                     filter_sql_expression=None, filter_requires_preprocessing=None, correlation_id=None,
                     message_id=None, to=None, reply_to=None, label=None, session_id=None, reply_to_session_id=None,
                     content_type=None, requires_preprocessing=None):

    if action_sql_expression:
        instance.action.sql_expression = action_sql_expression

    if action_compatibility_level:
        instance.action.compatibility_level = action_compatibility_level

    if action_requires_preprocessing:
        instance.action.requires_preprocessing = action_requires_preprocessing

    if filter_sql_expression:
        instance.sql_filter.sql_expression = filter_sql_expression

    if filter_requires_preprocessing:
        instance.sql_filter.requires_preprocessing = filter_requires_preprocessing

    if correlation_id:
        instance.correlation_filter.correlation_id = correlation_id

    if to:
        instance.correlation_filter.to = to

    if message_id:
        instance.correlation_filter.message_id = message_id

    if reply_to:
        instance.correlation_filter.reply_to = reply_to

    if label:
        instance.correlation_filter.label = label

    if session_id:
        instance.correlation_filter.session_id = session_id

    if reply_to_session_id:
        instance.correlation_filter.reply_to_session_id = reply_to_session_id

    if content_type:
        instance.correlation_filter.content_type = content_type

    if requires_preprocessing:
        instance.correlation_filter.requires_preprocessing = requires_preprocessing

    return instance


def cli_migration_start(client, resource_group_name, namespace_name, target_namespace, post_migration_name):
    import time

    client.create_and_start_migration(resource_group_name, namespace_name, target_namespace, post_migration_name)
    getresponse = client.get(resource_group_name, namespace_name)

    # pool till Provisioning state is succeeded
    while getresponse.provisioning_state != 'Succeeded':
        time.sleep(30)
        getresponse = client.get(resource_group_name, namespace_name)

    # poll on the 'pendingReplicationOperationsCount' to be 0 or none
    while getresponse.pending_replication_operations_count != 0 and getresponse.pending_replication_operations_count is not None:
        time.sleep(30)
        getresponse = client.get(resource_group_name, namespace_name)

    return client.get(resource_group_name, namespace_name)


iso8601pattern = re.compile("^P(?!$)(\\d+Y)?(\\d+M)?(\\d+W)?(\\d+D)?(T(?=\\d)(\\d+H)?(\\d+M)?(\\d+.)?(\\d+S)?)?$")
timedeltapattern = re.compile("^\\d+:\\d+:\\d+$")


def return_valid_duration(instance_value, update_value):
    from datetime import timedelta
    from isodate import parse_duration
    from azure.cli.command_modules.servicebus.constants import DURATION_SECS, DURATION_MIN, DURATION_DAYS
    if update_value is not None:
        value_toreturn = update_value
    else:
        value_toreturn = str(instance_value)

    if iso8601pattern.match(value_toreturn):
        if parse_duration(value_toreturn) > timedelta(days=DURATION_DAYS, minutes=DURATION_MIN, seconds=DURATION_SECS):
            return timedelta(days=DURATION_DAYS, minutes=DURATION_MIN, seconds=DURATION_SECS)
        return value_toreturn

    if timedeltapattern.match(value_toreturn):
        day, minute, seconds = value_toreturn.split(":")
        if timedelta(days=int(day), minutes=int(minute), seconds=int(seconds)) <= timedelta(days=DURATION_DAYS,
                                                                                            minutes=DURATION_MIN,
                                                                                            seconds=DURATION_SECS):
            return timedelta(days=int(day), minutes=int(minute), seconds=int(seconds))
        return timedelta(days=DURATION_DAYS, minutes=DURATION_MIN, seconds=DURATION_SECS)


# to check the timespan value
def return_valid_duration_create(update_value):
    from datetime import timedelta
    from isodate import parse_duration
    from knack.util import CLIError
    from azure.cli.command_modules.servicebus.constants import DURATION_SECS, DURATION_MIN, DURATION_DAYS
    if update_value is not None:
        if iso8601pattern.match(update_value):
            if parse_duration(update_value) > timedelta(days=DURATION_DAYS, minutes=DURATION_MIN, seconds=DURATION_SECS):
                raise CLIError(
                    'duration value should be less than (days:min:secs) 10675199:10085:477581')

        if timedeltapattern.match(update_value):
            day, minute, seconds = update_value.split(":")
            if timedelta(days=int(day), minutes=int(minute), seconds=int(seconds)) <= timedelta(days=DURATION_DAYS, minutes=DURATION_MIN, seconds=DURATION_SECS):
                update_value = timedelta(days=int(day), minutes=int(minute), seconds=int(seconds))
            else:
                raise CLIError(
                    'duration value should be less than timedelta(days=DURATION_DAYS, minutes=DURATION_MIN, seconds=DURATION_SECS)')

    return update_value


# NetwrokRuleSet Region
def cli_networkrule_createupdate(client, resource_group_name, namespace_name, subnet=None, ip_mask=None, ignore_missing_vnet_service_endpoint=False, action='Allow'):
    from azure.mgmt.servicebus.models import NWRuleSetVirtualNetworkRules, Subnet, NWRuleSetIpRules
    netwrokruleset = client.get_network_rule_set(resource_group_name, namespace_name)

    if netwrokruleset.virtual_network_rules is None:
        netwrokruleset.virtual_network_rules = []

    if netwrokruleset.ip_rules is None:
        netwrokruleset.ip_rules = []

    if subnet:
        netwrokruleset.virtual_network_rules.append(NWRuleSetVirtualNetworkRules(subnet=Subnet(id=subnet),
                                                                                 ignore_missing_vnet_service_endpoint=ignore_missing_vnet_service_endpoint))

    if ip_mask:
        netwrokruleset.ip_rules.append(NWRuleSetIpRules(ip_mask=ip_mask, action=action))

    return client.create_or_update_network_rule_set(resource_group_name, namespace_name, netwrokruleset)


def cli_networkrule_delete(client, resource_group_name, namespace_name, subnet=None, ip_mask=None):
    from azure.mgmt.servicebus.models import NWRuleSetVirtualNetworkRules, NWRuleSetIpRules
    netwrokruleset = client.get_network_rule_set(resource_group_name, namespace_name)

    if subnet:
        virtualnetworkrule = NWRuleSetVirtualNetworkRules()
        virtualnetworkrule.subnet = subnet

        for vnetruletodelete in netwrokruleset.virtual_network_rules:
            if vnetruletodelete.subnet.id.lower() == subnet.lower():
                virtualnetworkrule.ignore_missing_vnet_service_endpoint = vnetruletodelete.ignore_missing_vnet_service_endpoint
                netwrokruleset.virtual_network_rules.remove(vnetruletodelete)
                break

    if ip_mask:
        ipruletodelete = NWRuleSetIpRules()
        ipruletodelete.ip_mask = ip_mask
        ipruletodelete.action = "Allow"

        if ipruletodelete in netwrokruleset.ip_rules:
            netwrokruleset.ip_rules.remove(ipruletodelete)

    return client.create_or_update_network_rule_set(resource_group_name, namespace_name, netwrokruleset)


def cli_returnnsdetails(cmd, resource_group_name, namespace_name, max_size_in_megabytes):
    from knack.util import CLIError
    from azure.mgmt.servicebus import ServiceBusManagementClient
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    nsclient = get_mgmt_service_client(cmd.cli_ctx, ServiceBusManagementClient).namespaces
    getnamespace = nsclient.get(resource_group_name=resource_group_name, namespace_name=namespace_name)
    if getnamespace.sku.name == 'Standard' and max_size_in_megabytes not in [1024, 2048, 3072, 4096, 5120]:
        raise CLIError('--max-size on Standard sku namespace only supports upto [1024, 2048, 3072, 4096, 5120] GB')

    if getnamespace.sku.name == 'Premium' and max_size_in_megabytes not in [1024, 2048, 3072, 4096, 5120, 10240, 20480,
                                                                            40960, 81920]:
        raise CLIError(
            '--max-size on Premium sku namespace only supports upto [1024, 2048, 3072, 4096, 5120, 10240, 20480, 40960, 81920] GB')
